from src.tqdme.server import Server
from flask import send_from_directory
from pathlib import Path

script_directory = Path(__file__).parent.resolve()

if __name__ == "__main__":

    server = Server(script_directory, 'localhost', 3768)

    app = server.app

    @server.app.route('/src/<path:path>')
    def get_static_assets(path):
        return send_from_directory(server.base / 'src', path)

    server.run()


import json
from flask import Flask, request, jsonify, send_file
from flask_cors import cross_origin
from flask_socketio import SocketIO, join_room, leave_room

not_found_message = "<main style='padding: 25px'><h1>404 Error — No index.html found in the base path.</h1><p>Please provide one to visualize `tqdm` updates.</p></main>"

class Server:
    
    def __init__(self, base_path, host, port):
        self.base = base_path
        self.host = host
        self.port = port
        self.app, self.socketio = create(base_path, host, port)

    def run(self):
        self.socketio.run(self.app, host=self.host, port=self.port)

    def get_url(self, metadata):
        return get_url(self.host, self.port, metadata)

def get_url(host, port, metadata):
    user_id = metadata["user_id"]
    page_id = str(user_id)
    return f"http://{host}:{port}/view/{page_id}" 

def get_response(host, port, metadata):
    response = dict( ok = True )

    if metadata.get("url"):
        response["url"] = get_url(host, port, dict(user_id=metadata["user_id"]))

    return jsonify(response)

def get_page_id(metadata):
    return str(metadata['user_id'])

def update_states(states, metadata):
    page_id = get_page_id(metadata)
    identifier = f"{metadata['parent']}/{metadata['group']}/{metadata['id']}"

    changes = dict( user_id = None, id = None )

    if page_id not in states:
        states[page_id] = {}
        changes["user_id"] = True
        
    if identifier not in states[page_id]:
        states[page_id][identifier] = dict(done = False)
        changes["id"] = True

    state = states[page_id][identifier]
    state.update(metadata)

    if metadata.get("done"):
        changes["id"] = False

    return state, changes


def create(base_path, host, port):

    STATES = {}

    app = Flask(__name__)
    app.config['CORS_HEADERS'] = 'Content-Type'
    socketio = SocketIO(app, cors_allowed_origins="*")

    def update_local_state(metadata):

        metadata["user_id"] = request.remote_addr # Add request IP address as the unique User ID

        state, changes = update_states(STATES, metadata)

        user_changes = changes.get("user_id")
        if user_changes:
            url = get_url(host, port, metadata)
            message = 'onremoved' if not user_changes else 'onadded'
            socketio.emit(message, dict(id = get_page_id(metadata), url = url ))

        id_changes = changes.get("id")
        if id_changes is not None:
            message = 'onstart' if id_changes else 'onend'
            socketio.emit(message, dict(id = metadata['id']))

        return state

    @app.route('/')
    def index():
        try:
            return send_file(base_path / 'index.html')
        except:
            return not_found_message

    @app.route('/view/<path:path>')
    def view(path):
        try:
            return send_file(base_path / 'index.html')
        except:
            return not_found_message
        
    @app.route('/ping', methods=['POST'])
    @cross_origin()
    def ping():
        data = json.loads(request.data) if request.data else {}
        update_local_state(data)
        return get_response(host, port, data)

    @app.route('/update', methods=['POST'])
    @cross_origin()
    def update():
        data = json.loads(request.data) if request.data else {}

        state = update_local_state(data)
        
        # Send to frontend
        socketio.emit('progress', state, room=request.remote_addr)

        # Create pages for each User ID
        return get_response(host, port, data)

    
    @socketio.on('subscribe')
    def subscribe(page_id):
        user_id = page_id
        join_room(user_id) # Join room with User ID
        socketio.emit('init', dict(user_id=user_id, states=STATES.get(user_id, {}))) # Send initial state to client

    @socketio.on('unsubscribe')
    def unsubscribe(page_id):
        user_id = page_id
        leave_room(user_id) # Leave room with User ID


    @socketio.on('discover')
    def discover():
        user_ids = {}
        for user_id in STATES.keys():
            user_ids[user_id] = get_url(host, port, dict(user_id=user_id))
        socketio.emit('users', user_ids)

    return app, socketio



"""Demo of parallel tqdm visualization using tqdm.me polyfill"""

# Replace tqdm with tqdme (works in other packages)
from src.tqdme import tqdme
import tqdm
tqdm.tqdm = tqdme

# Load environment variables
from dotenv import load_dotenv
load_dotenv() 

# Import the necessary libraries for the demo
import time
from datetime import datetime
from typing import List
from concurrent.futures import ProcessPoolExecutor, as_completed
from threading import Thread

from tqdm import tqdm

N_JOBS = 3

# Each outer entry is a list of 'tasks' to perform on a particular worker
# For demonstration purposes, each in the list of tasks is the length of time in seconds
# that each iteration of the task takes to run and update the progress bar (emulated by sleeping)
BASE_SECONDS_PER_TASK = 0.5  # The base time for each task; actual time increases proportional to the index of the task
NUMBER_OF_TASKS_PER_JOB = 5
TASK_TIMES: List[List[float]] = [
    [BASE_SECONDS_PER_TASK * task_index] * NUMBER_OF_TASKS_PER_JOB
    for task_index in range(1, NUMBER_OF_TASKS_PER_JOB + 1)
]

def _run_sleep_tasks_in_subprocess(
    task_times: List[float],
    iteration_index: int,
    group: str
):
    """
    Run a 'task' that takes a certain amount of time to run on each worker.

    In this case that task is simply time.sleep.

    Parameters
    ----------
    sleep_time : float
        The amount of time this task emulates having taken to complete.
    iteration_index : int
        The index of this task in the list of all tasks from the buffer map.
        Each index would map to a different tqdm position.
    """

    sub_progress_bar = tqdm(
        iterable=task_times,
        position=iteration_index + 1,
        desc=f"Progress on iteration {iteration_index}",
        leave=False,
        tqdme_options=dict(
            parent=group
        )
    )

    for sleep_time in sub_progress_bar:
        time.sleep(sleep_time)


def run_parallel_processes(*, all_task_times: List[List[float]], n_jobs: int = 2):

    utc_now = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
    group = utc_now
    

    futures = list()
    with ProcessPoolExecutor(max_workers=n_jobs) as executor:

        # # Assign the parallel jobs
        for iteration_index, task_times_per_job in enumerate(all_task_times):
            futures.append(
                executor.submit(
                    _run_sleep_tasks_in_subprocess,
                    task_times=task_times_per_job,
                    iteration_index=iteration_index,
                    group=group
                )
            )

        total_tasks_iterable = as_completed(futures)
        total_tasks_progress_bar = tqdm(
            iterable=total_tasks_iterable, 
            total=len(all_task_times), 
            desc=f"Total tasks completed",
            tqdme_options=dict(
                group=group
            )
        )

        # Trigger the deployment of the parallel jobs
        for _ in total_tasks_progress_bar:
            pass

if __name__ == '__main__':
    run_parallel_processes(all_task_times=TASK_TIMES, n_jobs=N_JOBS)
    run_parallel_processes(all_task_times=TASK_TIMES, n_jobs=N_JOBS)




import os
from typing import Union
import json
from uuid import uuid4
import urllib3
from multiprocessing import Value
import io

from tqdm import tqdm as base_tqdm
from .utils import getBoolEnv

ACTIVE_BARS = dict()

DEFAULT_CONFIG = dict(
    verbose = lambda: getBoolEnv('TQDME_VERBOSE'),
    display = lambda: getBoolEnv('TQDME_DISPLAY'),
    url = lambda: os.getenv('TQDME_URL', 'http://tqdm.me'),
    id = lambda: str(uuid4()),
    group = lambda: os.getpid(),
    parent = lambda: os.getppid()
)

METADATA_TO_SEND = [ 'id', 'group', 'parent' ]

class tqdme(base_tqdm):

    __connected = Value('i', 1)

    __notifications = dict(
        connected = Value('i', 0),
        failure= Value('i', 0)
    )

    def __init__(
            self, 
            *args, 
            tqdme_options: dict = dict(),
            **kwargs
        ):

        config = tqdme_options.copy()
        # Resolve the configuration options for TQDME
        for key, func in DEFAULT_CONFIG.items():
            if key not in config:
                config[key] = func()

        self.__tqdme = config

        self.__done = False

        metadata = self.__sendrequest('ping', dict(url=True)) # Send a ping request to the server

        ACTIVE_BARS[self.__tqdme['id']] = self
        
        # If the server is connected, display the URL
        is_connected = self.__connected
        is_connected.acquire()
        is_currently_connected = is_connected.value
        is_connected.release()
        if is_currently_connected and metadata:
            url = metadata.get('url')
            if url:
                print(f"\nVisit {url} to view progress updates\n")

        else:
            failure_notification = self.__notifications['failure']
            failure_notification.acquire()
            if not failure_notification.value:
                print("\nFailed to connect to TQDME server\n")
                failure_notification.value = 1
            failure_notification.release()

        # Block display on the console
        if not self.__tqdme["display"] and 'file' not in kwargs:
            kwargs['file'] = BlockTqdmDisplay()

        # Initialize the base tqdm class
        super().__init__(*args, **kwargs)

        # Send initialization
        update = dict(format=self.format_dict.copy())
        self.__sendrequest('update', update)

    # Override the update method to run a callback function
    def update(self, n: int = 1) -> Union[bool, None]:
        displayed = super().update(n)

        self.__sendupdate()
        return displayed
    
    # Add cleanup method
    def __del__(self):
        self.cleanup()

    def cleanup(self):
        if not self.__done:
            self.__sendrequest('ping', dict(done=True))
            ACTIVE_BARS.pop(self.__tqdme['id'], None)
            self.__done = True

    # Always send a consistent update
    def __sendupdate(self):
        update = dict(format=self.format_dict.copy())
        if update['format']['n'] == update['format']['total']:
            update['done'] = True

        return self.__sendrequest('update', update)

    # Check if the server has been rejected
    def __isconnected(self):
        is_connected = self.__connected
        is_connected.acquire()
        to_request_url = is_connected.value == 1
        is_connected.release()
        return to_request_url
    
    # Send a request to the server
    def __sendrequest(self, pathname: str, data: dict = dict()):

        if not self.__isconnected():
            return

        url = f"{self.__tqdme['url']}/{pathname}" 

        http = urllib3.PoolManager()

        try:
            to_send = { key: self.__tqdme[key] for key in METADATA_TO_SEND }
            to_send.update(data)

            response = http.request('POST', url, body=json.dumps(to_send), headers={'Content-Type': 'application/json'})
            if response.status == 200:
                return json.loads(response.data)
            else:
                raise Exception(f"Failed to send POST request. Status code: {response.status}")

        except Exception as e:

            if self.__tqdme["verbose"]:
                print(f"An error occurred: {e}")

            is_connected = self.__connected
            is_connected.acquire()
            is_connected.value = 0
            is_connected.release()

class BlockTqdmDisplay(io.StringIO):
    def write(self, s):
        pass

    def flush(self):
        pass


# Ensure proper exit
import atexit
import signal

def exit_handler():
    for bar in list(ACTIVE_BARS.values()):
        bar.cleanup()

atexit.register(exit_handler)

def signal_handler(signum, frame):
    exit_handler()  # Call exit_handler before exiting
    raise SystemExit

signal.signal(signal.SIGTERM, signal_handler)
signal.signal(signal.SIGINT, signal_handler)
